#!/usr/bin/env python

"""
View annotated sources.

Copyright (C) 2006, 2007, 2010 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from micropython.common import *
from micropython.data import *
from os.path import exists, extsep, join
import sys
import os
import textwrap

# Classes.

# HTML-related output production.

html_header = """<?xml version="1.0" encoding="iso-8859-15"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
<head>
  <title>Module: %(full_name)s</title>
  <style type="text/css">
    html {
      background-color: black; color: white;
    }

    body {
      padding-bottom: 4em;
      font-size: 14pt; font-family: monospace;
      background-color: black; color: white;
    }

    a {
      text-decoration: none;
    }

    .nowrap { white-space: nowrap; }

    .class { margin-top: 1em; margin-bottom: 1em; }
    .function { margin-top: 1em; margin-bottom: 1em; }
    .body { padding-left: 2em; }
    .keyword { color: yellow; }
    .comment { color: blue; }
    .class-name { color: cyan; }
    .function-name { color: cyan; }
    .str { color: #FF00FF; }
    .doc { color: #FF00FF; margin-top: 1em; margin-bottom: 1em; }

    .popup {
      display: none;
      position: absolute;
      top: 3ex; left: 0;
      color: white;
      z-index: 3;
    }

    .scope { padding: 0.5em; background-color: #007700; }

    .name {
      position: relative;
      color: white;
    }

    .name:hover {
      background-color: #003300;
      padding-bottom: 1ex;
      z-index: 2;
    }

    .name:hover .popup {
      display: block;
    }

    .summary-class {
      background-color: #004400;
    }

    .summary-instance {
      background-color: #0000FF;
    }

    .summary-attr {
      background-color: #007700;
    }

  </style>
</head>
<body>
"""

html_footer = """</body>
</html>
"""

# Utility classes.

class Writer:

    "A utility class providing useful HTML output methods."

    # Methods which return strings.

    def _text(self, text):
        return text.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")

    def _attr(self, attr):
        return self._text(attr).replace("'", "&apos;").replace('"', "&quot;")

    def _url(self, url):
        return self._attr(url).replace("#", "%23").replace("-", "%2d")

    # Methods which write to the stream.

    def _span_end(self):
        self.stream.write("</span>")

    def _comment(self, comment):
        self.stream.write("<span class='comment'># %s</span>\n" % comment)

    def _reserved(self, token, classes, leading=0, trailing=1):
        if leading:
            self.stream.write(" ")
        self.stream.write("<span class='%s'>%s</span>" % (classes, self._text(token)))
        if trailing:
            self.stream.write(" ")

    def _keyword(self, kw, leading=0, trailing=1):
        self._reserved(kw, "keyword", leading, trailing)

    def _op(self, symbol, leading=0, trailing=1):
        self._reserved(symbol, "operator", leading, trailing)

    def _doc(self, node):
        if node.doc is not None:
            self.stream.write("<pre class='doc'>")
            self.stream.write('"""')
            output = textwrap.dedent(node.doc.replace('"""', '\\"\\"\\"'))
            self.stream.write(self._text(output))
            self.stream.write('"""')
            self.stream.write("</pre>\n")

    def _summary_link(self, module_name, full_name, name, classes=None):
        self._name_link("%s-summary" % module_name, full_name, name, classes)

    def _object_name_def(self, module, obj, classes=None):

        """
        Link to the summary for 'module' using 'obj'. The optional 'classes'
        can be used to customise the CSS classes employed.
        """

        self._summary_link(module.full_name(), obj.full_name(), obj.name, classes)

    def _name_start(self, classes=None):
        self.stream.write("<span class='%s'>" % (classes or "name"))

    _name_end = _span_end

    def _name(self, name, classes=None):
        self._name_start(classes)
        self.stream.write(name)
        self._name_end()

    def _name_link(self, module_name, full_name, name, classes=None):
        self.stream.write("<a class='%s' href='%s%sxhtml#%s'>%s</a>" % (
            classes or "name", module_name, os.path.extsep,
            self._attr(full_name), self._text(name)))

    def _popup_start(self):
        self.stream.write("<span class='popup'>")

    _popup_end = _span_end

    def _scope(self, scope):
        self.stream.write("<div class='scope'>%s</div>\n" % scope)

# Summary classes.

class Summary(Writer):

    "Summarise classes and attributes in modules."

    def __init__(self, module):
        self.module = module

    def to_stream(self, stream):

        "Write the summary to the given 'stream'."

        self.stream = stream
        self._init_details()
        self.stream.write(html_header % {"full_name" : self.module.full_name()})
        self._write_classes(self.module)
        self.stream.write(html_footer)

    def _write_classes(self, module):
        self.stream.write("<table cellspacing='5' cellpadding='5'>\n")

        all_classes = {}

        for obj in self.module.all_objects:
            if isinstance(obj, Class):
                all_classes[obj.name] = obj

        all_class_names = all_classes.keys()
        all_class_names.sort()

        for name in all_class_names:
            self._write_class(all_classes[name])

        self.stream.write("</table>\n")

    def _write_class(self, obj):

        # Write the class...

        self.stream.write("<tbody class='class'>\n")
        self.stream.write("<tr>\n")
        self.stream.write("<th class='summary-class' id='%s'>\n" % obj.full_name())
        self._keyword("class")
        self.stream.write(obj.name)
        self.stream.write("</th>\n")

        # ...and all known attribute names.

        obj_attributes = obj.all_attribute_names()

        for name in self.attribute_names:
            if name in obj_attributes:
                self.stream.write("<th class='summary-attr'>%s</th>\n" % self._text(name))
            else:
                self.stream.write("<th></th>\n")
        self.stream.write("</tr>\n")

        self.stream.write("</tbody>\n")

    def _init_details(self):
        names = set()

        # Visit all classes.

        for obj in self.module.all_objects:
            if isinstance(obj, Class):
                names.update(obj.all_attribute_names())

        self.attribute_names = list(names)
        self.attribute_names.sort()

# Source code classes.

class AnnotatedSource(ASTVisitor, Writer):

    "A module source code browser."

    def __init__(self, module):
        ASTVisitor.__init__(self)
        self.visitor = self
        self.module = module

    def to_stream(self, stream):

        "Write the annotated code to the given 'stream'."

        self.stream = stream
        self.stream.write(html_header % {"full_name" : self.module.full_name()})
        self.dispatch(self.module.astnode)
        self.stream.write(html_footer)

    def visitModule(self, node):
        self.default(node)

    # Statements.

    def visitAssert(self, node):
        self.stream.write("<div class='assert nowrap'>\n")
        self._keyword("assert")
        self.dispatch(node.test)
        if node.fail:
            self.stream.write(", ")
            self.dispatch(node.fail)
        self.stream.write("</div>\n")

    def visitAssign(self, node):
        self.stream.write("<div class='assign nowrap'>\n")
        for lvalue in node.nodes:
            self.dispatch(lvalue)
            self.stream.write(" = ")
        self.dispatch(node.expr)
        self.stream.write("</div>\n")

    def visitAugAssign(self, node):
        self.stream.write("<div class='augassign nowrap'>\n")
        self.dispatch(node.node)
        self.stream.write(" %s " % node.op)
        self.dispatch(node.expr)
        self.stream.write("</div>\n")

    def visitBreak(self, node):
        self.stream.write("<div class='break nowrap'>\n")
        self._keyword("break")
        self.stream.write("</div>\n")

    def visitClass(self, node):
        if not used_by_unit(node):
            return

        # Use inspected details where possible.

        if hasattr(node, "_def"):
            cls = node._def
            bases = cls.bases
            self.stream.write("<div class='class nowrap' id='%s'>\n" % cls.full_name())
        else:
            print "Warning: class %s not recognised!" % node.name
            return

        # Write the declaration line.

        self.stream.write("<div>\n")
        self._keyword("class")
        self._object_name_def(self.module, cls, "class-name")

        # Suppress the "object" class appearing alone.

        if bases and not (len(bases) == 1 and bases[0].name == "object"):
            self.stream.write("(")
            first = 1
            for base in bases:
                if not first:
                    self.stream.write(", ")

                self._object_name_def(base.module, base)

                first = 0
            self.stream.write(")")

        self.stream.write(":\n")
        self.stream.write("</div>\n")

        # Write the docstring and class body.

        self.stream.write("<div class='body nowrap'>\n")
        self._doc(node)
        self.dispatch(node.code)
        self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    def visitContinue(self, node):
        self.stream.write("<div class='continue nowrap'>\n")
        self._keyword("continue")
        self.stream.write("</div>\n")

    def visitDiscard(self, node):
        self.stream.write("<div class='discard nowrap'>\n")
        self.default(node)
        self.stream.write("</div>\n")

    def visitFor(self, node):
        self.stream.write("<div class='if nowrap'>\n")
        self.stream.write("<div>\n")
        self._keyword("for")
        self.dispatch(node.assign)
        self._keyword("in", 1)
        self.dispatch(node.list)
        self.stream.write(":\n")
        self.stream.write("</div>\n")
        self.stream.write("<div class='body nowrap'>\n")
        self.dispatch(node.body)
        self.stream.write("</div>\n")
        if node.else_ is not None:
            self.stream.write("<div>\n")
            self._keyword("else", trailing=0)
            self.stream.write(":\n")
            self.stream.write("</div>\n")
            self.stream.write("<div class='body nowrap'>\n")
            self.dispatch(node.else_)
            self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    def visitFrom(self, node):
        self.stream.write("<div class='from nowrap'>\n")
        self._keyword("from")
        self._name(node.modname)
        self._keyword("import", 1)
        first = 1
        for (name, alias), _name in map(None, node.names, node._names):
            if not first:
                self.stream.write(", ")
            if alias:
                self.stream.write(name + " ")
                self._keyword("as", 1)
            self._name(alias or name)
            first = 0
        self.stream.write("</div>\n")

    def visitFunction(self, node):
        if not used_by_unit(node):
            return

        if hasattr(node, "_def"):
            fn = node._def
            self.stream.write("<div class='function nowrap' id='%s'>\n" % fn.full_name())
        else:
            print "Warning: function %s not recognised!" % node.name
            return

        # Write the declaration line.

        self.stream.write("<div>\n")
        self._keyword("def")
        self._object_name_def(self.module, fn, "function-name")

        self.stream.write("(")
        self._parameters(fn)
        self.stream.write(")")
        self.stream.write(":\n")
        self.stream.write("</div>\n")

        self.stream.write("<div class='body nowrap'>\n")
        self._doc(node)
        self.dispatch(node.code)
        self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    def visitGlobal(self, node):
        self.stream.write("<div class='global nowrap'>\n")
        self._keyword("global")
        first = 1
        for name in node.names:
            if not first:
                self.stream.write(", ")
            self.stream.write(name)
            first = 0
        self.stream.write("</div>\n")

    def visitIf(self, node):
        self.stream.write("<div class='if nowrap'>\n")
        first = 1
        for compare, stmt in node.tests:
            self.stream.write("<div>\n")
            if first:
                self._keyword("if")
            else:
                self._keyword("elif")
            self.dispatch(compare)
            self.stream.write(":\n")
            self.stream.write("</div>\n")
            self.stream.write("<div class='body nowrap'>\n")
            self.dispatch(stmt)
            self.stream.write("</div>\n")
            first = 0
        if node.else_ is not None:
            self.stream.write("<div>\n")
            self._keyword("else", trailing=0)
            self.stream.write(":\n")
            self.stream.write("</div>\n")
            self.stream.write("<div class='body nowrap'>\n")
            self.dispatch(node.else_)
            self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    def visitImport(self, node):
        self.stream.write("<div class='import nowrap'>\n")
        self._keyword("import")
        first = 1
        for name, alias in node.names:
            if not first:
                self.stream.write(",\n")
            if alias:
                self.stream.write(name + " ")
                self._keyword("as", 1)
            self._name(alias or name)
            first = 0
        self.stream.write("</div>\n")

    def visitPass(self, node):
        self.stream.write("<div class='pass nowrap'>\n")
        self._keyword("pass")
        self.stream.write("</div>\n")

    def visitPrint(self, node):
        self.stream.write("<div class='print nowrap'>\n")
        self._keyword("print")
        if node.dest is not None:
            self.stream.write(">>\n")
            self.dispatch(node.dest)
        for n in node.nodes:
            self.dispatch(n)
            self.stream.write(",\n")
        self.stream.write("</div>\n")

    def visitPrintnl(self, node):
        self.stream.write("<div class='printnl nowrap'>\n")
        self._keyword("print")
        if node.dest is not None:
            self.stream.write(">>\n")
            self.dispatch(node.dest)
        first = 1
        for n in node.nodes:
            if not first:
                self.stream.write(",\n")
            self.dispatch(n)
            first = 0
        self.stream.write("</div>\n")

    def visitRaise(self, node):
        self.stream.write("<div class='raise nowrap'>\n")
        self._keyword("raise")
        self.dispatch(node.expr1)
        if node.expr2 is not None:
            self.stream.write(",\n")
            self.dispatch(node.expr2)
        if node.expr3 is not None:
            self.stream.write(",\n")
            self.dispatch(node.expr3)
        self.stream.write("</div>\n")

    def visitReturn(self, node):
        self.stream.write("<div class='return nowrap'>\n")
        self._keyword("return")
        self.dispatch(node.value)
        self.stream.write("</div>\n")

    def visitStmt(self, node):
        self.stream.write("<div class='stmt nowrap'>\n")
        self.default(node)
        self.stream.write("</div>\n")

    def visitTryExcept(self, node):
        self.stream.write("<div class='tryexcept nowrap'>\n")
        self.stream.write("<div>\n")
        self._keyword("try", trailing=0)
        self.stream.write(":\n")
        self.stream.write("</div>\n")
        self.stream.write("<div class='body nowrap'>\n")
        self.dispatch(node.body)
        self.stream.write("</div>\n")
        for spec, assign, statement in node.handlers:
            self.stream.write("<div>\n")
            self._keyword("except")
            if spec is not None:
                self.dispatch(spec)
            if assign is not None:
                self.stream.write(",\n")
                self.dispatch(assign)
            self.stream.write(":\n")
            self.stream.write("</div>\n")
            self.stream.write("<div class='body nowrap'>\n")
            self.dispatch(statement)
            self.stream.write("</div>\n")
        if node.else_ is not None:
            self.stream.write("<div>\n")
            self._keyword("else", trailing=0)
            self.stream.write(":\n")
            self.stream.write("</div>\n")
            self.stream.write("<div class='body nowrap'>\n")
            self.dispatch(node.else_)
            self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    def visitTryFinally(self, node):
        self.stream.write("<div class='tryfinally nowrap'>\n")
        self.stream.write("<div>\n")
        self._keyword("try", trailing=0)
        self.stream.write(":\n")
        self.stream.write("</div>\n")
        self.stream.write("<div class='body nowrap'>\n")
        self.dispatch(node.body)
        self.stream.write("</div>\n")
        self.stream.write("<div>\n")
        self._keyword("finally", trailing=0)
        self.stream.write(":\n")
        self.stream.write("</div>\n")
        self.stream.write("<div class='body nowrap'>\n")
        self.dispatch(node.final)
        self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    def visitWhile(self, node):
        self.stream.write("<div class='while nowrap'>\n")
        self.stream.write("<div>\n")
        self._keyword("while")
        self.dispatch(node.test)
        self.stream.write(":\n")
        self.stream.write("</div>\n")
        self.stream.write("<div class='body nowrap'>\n")
        self.dispatch(node.body)
        self.stream.write("</div>\n")
        if node.else_ is not None:
            self.stream.write("<div>\n")
            self._keyword("else", trailing=0)
            self.stream.write(":\n")
            self.stream.write("</div>\n")
            self.stream.write("<div class='body nowrap'>\n")
            self.dispatch(node.else_)
            self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    # Expression-related helper methods.

    def _visitBitBinary(self, node, name, symbol):
        self.stream.write("<span class='%s'>" % name)
        first = 1
        for node in node.nodes:
            if not first:
                self._op(symbol, 1)
            self.dispatch(node)
            first = 0
        self.stream.write("</span>")

    def _visitBinary(self, node, name, symbol):
        self.stream.write("<span class='%s'>" % name)
        self.dispatch(node.left)
        self._op(symbol, 1)
        self.dispatch(node.right)
        self.stream.write("</span>")

    def _visitUnary(self, node, name, symbol):
        self.stream.write("<span class='%s'>" % name)
        self._op(symbol)
        self.dispatch(node.expr)
        self.stream.write("</span>")

    # Expressions.

    def visitAdd(self, node):
        self._visitBinary(node, "add", "+")

    def visitAnd(self, node):
        self.stream.write("<span class='and'>")
        first = 1
        for n in node.nodes:
            if not first:
                self._keyword("and", 1)
            self.dispatch(n)
            first = 0
        self.stream.write("</span>")

    def visitAssAttr(self, node):
        self.stream.write("<span class='assattr'>")
        self.dispatch(node.expr)
        self.stream.write("<span class='attr'>")
        self.stream.write(".")
        self._name(node.attrname)
        self.stream.write("</span>")
        self.stream.write("</span>")

    def visitAssList(self, node):
        self.stream.write("<span class='list'>")
        self.stream.write("[")
        self._sequence(node)
        self.stream.write("]")
        self.stream.write("</span>")

    def visitAssName(self, node):
        self._name(node.name)

    def visitAssTuple(self, node):
        self.stream.write("<span class='tuple'>")
        self.stream.write("(")
        self._sequence(node)
        self.stream.write(")")
        self.stream.write("</span>")

    def visitBitand(self, node):
        self._visitBitBinary(node, "bitand", "&")

    def visitCallFunc(self, node):
        self.stream.write("<span class='callfunc'>")
        self.dispatch(node.node)
        self.stream.write("<span class='call'>")
        self.stream.write("(")
        first = 1
        for arg in node.args:
            if not first:
                self.stream.write(", ")
            self.dispatch(arg)
            first = 0
        if node.star_args is not None:
            if not first:
                self.stream.write(", *")
            self.dispatch(node.star_args)
            first = 0
        if node.dstar_args is not None:
            if not first:
                self.stream.write(", **")
            self.dispatch(node.dstar_args)
            first = 0
        self.stream.write(")")
        self.stream.write("</span>")
        self.stream.write("</span>")

    def visitCompare(self, node):
        self.stream.write("<span class='compare'>")
        self.dispatch(node.expr)
        for op_name, expr in node.ops:
            self._op(op_name, 1)
            self.dispatch(expr)
        self.stream.write("</span>")

    def visitConst(self, node):
        if isinstance(node.value, (str, unicode)):
            self.stream.write("<span class='str'>")
        self.stream.write(self._text(repr(node.value)))
        if isinstance(node.value, (str, unicode)):
            self.stream.write("</span>")

    def visitDict(self, node):
        self.stream.write("<span class='dict'>")
        self.stream.write("{")
        self._mapping(node)
        self.stream.write("}")
        self.stream.write("</span>")

    def visitDiv(self, node):
        self._visitBinary(node, "div", "/")

    def visitFloorDiv(self, node):
        self._visitBinary(node, "floordiv", "//")

    def visitGetattr(self, node):
        self.stream.write("<span class='getattr'>")
        self.dispatch(node.expr)
        self.stream.write("<span class='attr'>")
        self.stream.write(".")
        self._name(node.attrname)
        self.stream.write("</span>")
        self.stream.write("</span>")

    def visitKeyword(self, node):
        self.stream.write("<span class='keyword-arg'>")
        self.stream.write(node.name)
        self.stream.write("=")
        self.dispatch(node.expr)
        self.stream.write("</span>")

    def visitLambda(self, node):
        if hasattr(node, "_def"):
            fn = node._def
        else:
            print "Warning: function %s not recognised!" % node.name
            return

        self.stream.write("<span class='lambda'>")
        self._keyword("lambda")
        self._parameters(fn)
        self.stream.write(": <span class='code'>")
        self.dispatch(node.code)
        self.stream.write("</span>")
        self.stream.write("</span>")

    visitList = visitAssList

    def visitListComp(self, node):
        self.stream.write("<span class='listcomp'>")
        self.stream.write("[")
        self.dispatch(node.expr)
        for qual in node.quals:
            self.dispatch(qual)
        self.stream.write("]")
        self.stream.write("</span>")

    def visitListCompFor(self, node):
        self.stream.write("<span class='listcompfor'>")
        self._keyword("for")
        self.stream.write("<span class='iterator'>")
        self.dispatch(node.assign)
        self.stream.write("</span>")
        self._keyword("in", 1)
        self.stream.write("<span class='collection'>")
        self.dispatch(node.list)
        self.stream.write("</span>")
        for if_ in node.ifs:
            self.dispatch(if_)
        self.stream.write("</span>")

    def visitListCompIf(self, node):
        self.stream.write("<span class='listcompif'>")
        self.stream.write("<span class='conditional'>")
        self._keyword("if", 1)
        self.dispatch(node.test)
        self.stream.write("</span>")
        self.stream.write("</span>")

    def visitMod(self, node):
        self._visitBinary(node, "mod", "%")

    def visitMul(self, node):
        self._visitBinary(node, "mul", "*")

    def visitName(self, node):
        if hasattr(node, "_scope"):
            scope = node._scope
            self._name_start()
            self.stream.write(node.name)
            self._popup_start()
            self._scope(scope)
            self._popup_end()
            self._name_end()
        else:
            self._name(node.name)

    # Output preparation methods.

    def _sequence(self, node):
        first = 1
        for n in node.nodes:
            if not first:
                self.stream.write(", ")
            self.dispatch(n)
            first = 0

    def _mapping(self, node):
        first = 1
        for k, v in node.items:
            if not first:
                self.stream.write(", ")
            self.dispatch(k)
            self.stream.write(" : ")
            self.dispatch(v)
            first = 0

    def _parameters(self, fn):
        nparams = len(fn.positional_names)
        ndefaults = len(fn.defaults)
        first_with_default = nparams - ndefaults

        first = 1
        for n, param in enumerate(fn.positional_names):
            if not first:
                self.stream.write(", ")
            self._name(param)
            n_default = n - first_with_default
            if n_default >= 0:
                self._default(fn.defaults[n_default])
            first = 0

        if fn.has_star:
            if not first:
                self.stream.write(", *")
            self._name(fn.star_name)

        if fn.has_dstar:
            if not first:
                self.stream.write(", **")
            self._name(fn.dstar_name)

    def _default(self, default):
        self.stream.write("=")
        self.dispatch(default)

# Convenience functions.

def summarise(module, filename):
    stream = open(filename, "wb")
    try:
        summary = Summary(module)
        summary.to_stream(stream)
    finally:
        stream.close()

def annotate(module, filename):
    stream = open(filename, "wb")
    try:
        source = AnnotatedSource(module)
        source.to_stream(stream)
    finally:
        stream.close()

def report(importer, directory):
    if not exists(directory):
        os.mkdir(directory)

    for module in importer.get_modules():
        annotate(module, join(directory, "%s%sxhtml" % (module.full_name(), extsep)))
        summarise(module, join(directory, "%s-summary%sxhtml" % (module.full_name(), extsep)))

# vim: tabstop=4 expandtab shiftwidth=4
